/*
	go-kobo-input - Basic touch input handling for Kobo ereaders
    Copyright (C) 2018 Sherman Perry

    This file is part of go-kobo-input.

    go-kobo-input is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    go-kobo-input is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with go-kobo-input.  If not, see <https://www.gnu.org/licenses/>.
*/

package main

import (
	"encoding/json"
	"fmt"
	"io/ioutil"
	"os"
	"syscall"
	"unsafe"
	"time"
	"errors"
	"strconv"
	"github.com/shermp/go-fbink-v2/gofbink"
	"github.com/shermp/go-kobo-input/koboin"
	"github.com/shermp/go-osk/osk"
)

var err error
var eno syscall.Errno
var ttyFile *os.File
var	cm ConfMap
var haveConf bool


type ConfMap struct {
	NumSymLayouts     int      `json:"numSymLayouts"`
	NumSymLayoutFonts int      `json:"numSymLayoutFonts"`
	SymLayoutFont     []string `json:"symlayoutFont"`
	NumProgKeys       int      `json:"numProgKeys"`
	ProgKeys          []struct {
		ProgKeyType     int    `json:"progKeyType"`
		ProgKeyLabel    string `json:"progKeyLabel"`
		ProgKeySeq      string `json:"progKeySeq"`
	} `json:"progKeys"`
}

// sendAnsiSeq sends an ANSI escape string to open tty  
func sendAnsiSeq(ansiSeq string) syscall.Errno {
	seqBytes := []byte(ansiSeq)
	for i := 0; i < len(seqBytes); i++ { 
		_, _, eno = syscall.Syscall(syscall.SYS_IOCTL,ttyFile.Fd(),syscall.TIOCSTI,uintptr(unsafe.Pointer(&seqBytes[i])),)
		if eno != 0 {
			return eno
		} 
	}
	return 0
}

func main() {

	// Create and init FBInk
	cfg := gofbink.FBInkConfig{}
	rCfg := gofbink.RestrictedConfig{}
	rCfg.Fontmult = 3
	rCfg.Fontname = gofbink.IBM
	fb := gofbink.New(&cfg, &rCfg)
	fb.Open()
	defer fb.Close()
	fb.Init(&cfg)
	fbState := gofbink.FBInkState{}
	fb.GetState(&cfg, &fbState)

	fb.Println("Starting Oskansi Keyboard...")

	// Output messages to fb if OSK_VERBOSE set from env
	verbose := false
	envstr := os.Getenv("OSK_VERBOSE")
	fmt.Println("OSK_VERBOSE is: ", envstr)
	if envstr != "" {
		verbose = true
	} else {
		verbose = false
	}

	// Load a conf file for the OSK
	conf1JSON, err := ioutil.ReadFile("./osk-keymap-conf.json")
	if err != nil {
		fmt.Println("Read of osk-keymap-conf.json failed, err: ", err)
		haveConf = false
		fmt.Println("haveConf = ", haveConf)
		if verbose {
			fb.Println(err)
			fb.Println("haveConf = ", haveConf)
		}
	} else {
		cm = ConfMap{}
		json.Unmarshal(conf1JSON, &cm)
		haveConf = true
		fmt.Println("haveConf = ", haveConf)
		fmt.Println("NumSymLayouts = ", cm.NumSymLayouts)
		fmt.Println("NumProgKeys = ", cm.NumProgKeys)
		if verbose {
			fb.Println("haveConf = ", haveConf)
			fb.Println("NumProgKeys = ", cm.NumProgKeys)
		}
	}


	// Get the input event filepath
	touchPath := "/dev/input/event1"
	// Create an input object 
	envstr = os.Getenv("OSK_WIDTH")
	fmt.Println("OS_WIDTH is: ", envstr)
	var width, _ = strconv.Atoi(envstr)
	if !((width > 100) && (width < 10000)) {
	 	width = int(fbState.ScreenWidth)
	 }
	envstr = os.Getenv("OSK_HEIGHT")
	fmt.Println("OS_HEIGHT is: ", envstr)
	var height, _ = strconv.Atoi(envstr)	
	if !((height > 100) && (height < 10000)) {
	 	height = int(fbState.ScreenHeight)
	 }
	t := koboin.New(touchPath, width, height)
	if t == nil {
		return
	}
	defer t.Close()
	fmt.Println("ScreenWidth is: ", width)
	fmt.Println("ScreenHeight is: ", height)
		if verbose {
			fb.Println("ScreenWidth is: ", width)
			fb.Println("ScreenHeight is: ", height)
		}

	var tty = os.Getenv("OSK_TTY")
	fmt.Println("OSK_TTY is: ", tty)
	ttyFile, err = os.Open(tty)
	if err != nil {
		fmt.Println(err)
		tty = "/dev/tty1"
		ttyFile, err = os.Open(tty)
		if err != nil {
			fmt.Println(err)
			tty = "/dev/pts/0"
			ttyFile, err = os.Open(tty)
			if err != nil {
				fmt.Println(err)
			}
		}
	}
	defer ttyFile.Close()


	// Get environment variables for layouts
	envstr = os.Getenv("OSK_REGEN")
	fmt.Println("OSK_REGEN is: ", envstr)
	var regen, _ = strconv.Atoi(envstr)
	if (regen > 0) {
		regen = 1
	} else {
	 	regen = 0
	}
	

	// Load a keymap file for LowerCase layout
	keymap1JSON, _ := ioutil.ReadFile("./osk-keymap-LC.json")
	km1 := osk.KeyMap{}
	json.Unmarshal(keymap1JSON, &km1)
	// Create an OSK
	vk1, _ := osk.New(&km1, width, height)
	// Generate an image of the OSK
	vk1PNG := "./osk-keymap-LC.png"
	vk1Font := os.Getenv("OSK_FONTMAIN")
	fmt.Println("OSK_FONTMAIN is: ", vk1Font)
	_, err = os.Stat(vk1Font)
	if err != nil {
		vk1Font = "./Roboto-Medium.ttf"
	}
	fmt.Println("Using font: ", vk1Font)
	_, err = os.Stat(vk1PNG)
	if (errors.Is(err, os.ErrNotExist) || regen > 0) {
		vk1.CreateIMG(vk1PNG, vk1Font)
	}

	// Load a keymap file for UpperCase layout
	keymap2JSON, _ := ioutil.ReadFile("./osk-keymap-UC.json")
	km2 := osk.KeyMap{}
	json.Unmarshal(keymap2JSON, &km2)
	// Create an OSK
	vk2, _ := osk.New(&km2, width, height)
	// Generate an image of the OSK
	vk2PNG := "./osk-keymap-UC.png"
	vk2Font := vk1Font
	_, err = os.Stat(vk2PNG)
	if (errors.Is(err, os.ErrNotExist) || regen > 0) {
		vk2.CreateIMG(vk2PNG, vk2Font)
	}

	
	numSymLayouts := 0
	numSymLayoutFonts := 0
	if !haveConf {
		_, err = os.Stat("./osk-keymap-S0.json")
		if err != nil {
			fmt.Println("Read of osk-keymap-S0.json failed, err: ", err)
		} else {
			numSymLayouts = 1
			numSymLayoutFonts = 0
		}
	} else {
		numSymLayouts = cm.NumSymLayouts
		numSymLayoutFonts = cm.NumSymLayoutFonts
	}
	fmt.Println("numSymLayouts = ", numSymLayouts)
	fmt.Println("numSymLayoutFonts = ", numSymLayoutFonts)
	if verbose {
		fb.Println("numSymLayouts = ", numSymLayouts)
		fb.Println("numSymLayoutFonts = ", numSymLayoutFonts)
	}

	var keyLayout = make([]string, numSymLayouts)
	var imgLayout = make([]string, numSymLayouts)
	var	symNState = make([]bool, numSymLayouts)
	kmS := osk.KeyMap{}
	vkS := make([]*osk.VirtKeyboard, numSymLayouts)
	for i := 0; i < numSymLayouts; i++ { 
		fmt.Println("Symbol Layout: ",i)
		if verbose {
			fb.Println("Symbol Layout: ",i)
		}
		// Load a keymap file for the OSK
		keyLayout[i] = "./osk-keymap-S" + strconv.Itoa(i) + ".json"
		keymapSJSON, _ := ioutil.ReadFile(keyLayout[i])
		json.Unmarshal(keymapSJSON, &kmS)
		// Create an OSK
		vkS[i], _ = osk.New(&kmS, width, height)
		// Generate an image of the OSK
		imgLayout[i] = "./osk-keymap-S" + strconv.Itoa(i) + ".png"
		_, err = os.Stat(imgLayout[i])
		if (errors.Is(err, os.ErrNotExist) || regen > 0) {
			if i < numSymLayoutFonts {
				vkS[i].CreateIMG(imgLayout[i], cm.SymLayoutFont[i])
			} else if numSymLayoutFonts > 0 {
				vkS[i].CreateIMG(imgLayout[i], cm.SymLayoutFont[cm.NumSymLayoutFonts - 1])
			} else {
				vkS[i].CreateIMG(imgLayout[i], vk1Font)
			}
		}
        symNState[i] = false
	}

	// Print the image to the screen. Its position on screen should match that stored
	// in the keyboard object
	fb.PrintImage(vk1PNG, int16(vk1.StartCoords.X), int16(vk1.StartCoords.Y), &cfg)

	var k osk.Key
	var	upperCase bool
	var	ctlState bool
	var	altState bool
	var	symState bool
	var prevType int
	var debounceStartTm time.Time
	upperCase = false
	ctlState = false
	altState = false
	symState = false

	// Read the input from the touch screen
L:
	for {
		x, y, err := t.GetInput()
		if err != nil {
			continue
		}
		if upperCase {
			k, err = vk2.GetPressedKey(x, y)
		} else if symState {
			for i := 0; i < numSymLayouts; i++ { 
				if symNState[i] {
					k, err = vkS[i].GetPressedKey(x, y)
				}
			}	
		} else {
			k, err = vk1.GetPressedKey(x, y)
		}
		if err != nil {
			continue
		}
		if !k.IsKey {
			continue
		}
		if (prevType != osk.KTcapsLock && prevType != osk.KTsym && (prevType  < osk.KTs0 || prevType >= osk.MaxKeyCodes) || (time.Since(debounceStartTm) > (500 * time.Millisecond))) {
			switch {
			case (k.KeyType == osk.KTstandardChar):
				key := k.KeyCode
				tabSeq := []byte(string(key))
				if ( ctlState && (len(tabSeq) == 1) && ( ((tabSeq[0] > 64) && (tabSeq[0] < 91)) || ((tabSeq[0] > 96) && (tabSeq[0] < 123)) ) ) {
					if ((tabSeq[0] > 64) && (tabSeq[0] < 91)) {
						tabSeq[0] -= 64
					} else if ((tabSeq[0] >96) && (tabSeq[0] < 123)) {
						tabSeq[0] -= 96
					}
					if altState {
						altSeq := []byte("\u001b")
						tabSeq = append(altSeq, tabSeq...)
					}
					seqErr := sendAnsiSeq(string(tabSeq)) 
					if seqErr != 0 {
						fmt.Println(seqErr)
					}
				} else {
					if altState {
						altSeq := []byte("\u001b")
						tabSeq = append(altSeq, tabSeq...)
					}
					seqErr := sendAnsiSeq(string(tabSeq)) 
					if seqErr != 0 {
						fmt.Println(seqErr)
					}
				}
				altState = false
				ctlState = false
			case (k.KeyType == osk.KTcarriageReturn):
				tabSeq := []byte{13}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTbackspace):
				tabSeq := []byte{8}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTdelete):
				tabSeq := []byte{127}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTcapsLock):
				upperCase = !upperCase
				if upperCase {
					symState = false
					for i := 0; i < numSymLayouts; i++ { 
						symNState[i] = false
					}	
					fb.PrintImage(vk2PNG, int16(vk2.StartCoords.X), int16(vk2.StartCoords.Y), &cfg)
				} else {
					fb.PrintImage(vk1PNG, int16(vk1.StartCoords.X), int16(vk1.StartCoords.Y), &cfg)
				}
				debounceStartTm = time.Now()
			case (k.KeyType == osk.KTcontrol):
				ctlState = true
			case (k.KeyType == osk.KTalt):
				altState = true
			case (k.KeyType == osk.KTesc):
				tabSeq := []byte{27}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTtab):
				if (upperCase) {
					tabSeq := []byte{27,91,90}
					seqErr := sendAnsiSeq(string(tabSeq)) 
					if seqErr != 0 {
						fmt.Println(seqErr)
					}
				} else {
					tabSeq := []byte{9}
					seqErr := sendAnsiSeq(string(tabSeq)) 
					if seqErr != 0 {
						fmt.Println(seqErr)
					}
				}
			case (k.KeyType == osk.KTup):
				tabSeq := []byte{27,91,65}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTdown):
				tabSeq := []byte{27,91,66}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTleft):
				tabSeq := []byte{27,91,68}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTright):
				tabSeq := []byte{27,91,67}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTpgUp):
				tabSeq := []byte{27,91,53,126}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTpgDn):
				tabSeq := []byte{27,91,54,126}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTinsert):
				tabSeq := []byte{27,91,50,126}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KThome):
				tabSeq := []byte{27,91,49,126}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTend):
				tabSeq := []byte{27,91,52,126}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTsym):
				symState = !symState
				for i := 0; i < numSymLayouts; i++ { 
					symNState[i] = false
				}	
				if symState {
					upperCase = false
					symNState[0] = true
					fb.PrintImage(imgLayout[0], int16(vkS[0].StartCoords.X), int16(vkS[0].StartCoords.Y), &cfg)
				} else {
					fb.PrintImage(vk1PNG, int16(vk1.StartCoords.X), int16(vk1.StartCoords.Y), &cfg)
				}
				debounceStartTm = time.Now()
			case (k.KeyType == osk.KTf1):
				tabSeq := []byte{27,91,49,49,126}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTf2):
				tabSeq := []byte{27,91,49,50,126}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTf3):
				tabSeq := []byte{27,91,49,51,126}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTf4):
				tabSeq := []byte{27,91,49,52,126}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTf5):
				tabSeq := []byte{27,91,49,53,126}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTf6):
				tabSeq := []byte{27,91,49,54,126}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTf7):
				tabSeq := []byte{27,91,49,55,126}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTf8):
				tabSeq := []byte{27,91,49,56,126}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTf9):
				tabSeq := []byte{27,91,50,48,126}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTf10):
				tabSeq := []byte{27,91,50,49,126}
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType >= osk.KTs0 && k.KeyType < (osk.KTs0 + numSymLayouts) && k.KeyType <= osk.KTs99):
				for i := 0; i < numSymLayouts; i++ { 
					symNState[i] = false
				}	
				symNState[k.KeyType - osk.KTs0] = true
				fb.PrintImage(imgLayout[k.KeyType - osk.KTs0], int16(vkS[k.KeyType - osk.KTs0].StartCoords.X), int16(vkS[k.KeyType - osk.KTs0].StartCoords.Y), &cfg)
				debounceStartTm = time.Now()
			case (k.KeyType >= osk.MaxKeyCodes && k.KeyType < (osk.MaxKeyCodes + cm.NumProgKeys) && symState && !symNState[0]):
				tabSeq := []byte(cm.ProgKeys[k.KeyType - osk.MaxKeyCodes].ProgKeySeq)
				seqErr := sendAnsiSeq(string(tabSeq)) 
				if seqErr != 0 {
					fmt.Println(seqErr)
				}
			case (k.KeyType == osk.KTexit):
				break L
			default:
				continue
			}
			prevType = k.KeyType
		}
	}
}
