/*
	go-osk - Simple on-screen keyboard aimed at Kobo ereaders
    Copyright (C) 2018 Sherman Perry

    This file is part of go-osk.

    go-osk is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    go-osk is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with go-osk.  If not, see <https://www.gnu.org/licenses/>.
*/

package osk

import (
	"errors"
	"fmt"
	"math"
//	"strings"
	"time"

	"github.com/fogleman/gg"
)

const MaxKeyCodes = 1000

const (
	KTstandardChar   = 0
	KTcarriageReturn = 1
	KTbackspace      = 2
	KTdelete         = 3
	KTcapsLock       = 4
	KTcontrol        = 5
	KTalt            = 6
	KTesc            = 7
	KTtab            = 8
	KTup             = 9
	KTdown           = 10
	KTleft           = 11
	KTright          = 12
	KTpgUp           = 13
	KTpgDn           = 14
	KTinsert         = 15
	KThome           = 16
	KTend            = 17
	KTsym            = 18
	KTexit           = 19
	KTnop20          = 20
	KTnop21          = 21
	KTnop22          = 22
	KTnop23          = 23
	KTnop24          = 24
	KTnop25          = 25
	KTnop26          = 26
	KTnop27          = 27
	KTnop28          = 28
	KTnop29          = 29
	KTnop30          = 30
	KTnop31          = 31
	KTnop32          = 32
	KTnop33          = 33
	KTnop34          = 34
	KTnop35          = 35
	KTnop36          = 36
	KTnop37          = 37
	KTnop38          = 38
	KTnop39          = 39
	KTnop40          = 40
	KTnop41          = 41
	KTnop42          = 42
	KTnop43          = 43
	KTnop44          = 44
	KTnop45          = 45
	KTnop46          = 46
	KTnop47          = 47
	KTnop48          = 48
	KTnop49          = 49
	KTf1             = 50
	KTf2             = 51
	KTf3             = 52
	KTf4             = 53
	KTf5             = 54
	KTf6             = 55
	KTf7             = 56
	KTf8             = 57
	KTf9             = 58
	KTf10            = 59
	KTf11            = 60
	KTf12            = 61
	KTf13            = 62
	KTf14            = 63
	KTf15            = 64
	KTf16            = 65
	KTf17            = 66
	KTf18            = 67
	KTf19            = 68
	KTf20            = 69
	KTnop70          = 70
	KTnop71          = 71
	KTnop72          = 72
	KTnop73          = 73
	KTnop74          = 74
	KTnop75          = 75
	KTnop76          = 76
	KTnop77          = 77
	KTnop78          = 78
	KTnop79          = 79
	KTnop80          = 80
	KTnop81          = 81
	KTnop82          = 82
	KTnop83          = 83
	KTnop84          = 84
	KTnop85          = 85
	KTnop86          = 86
	KTnop87          = 87
	KTnop88          = 88
	KTnop89          = 89
	KTnop90          = 90
	KTnop91          = 91
	KTnop92          = 92
	KTnop93          = 93
	KTnop94          = 94
	KTnop95          = 95
	KTnop96          = 96
	KTnop97          = 97
	KTnop98          = 98
	KTnop99          = 99
	KTs0             = 100
	KTs1             = 101
	KTs2             = 102
	KTs3             = 103
	KTs4             = 104
	KTs5             = 105
	KTs6             = 106
	KTs7             = 107
	KTs8             = 108
	KTs9             = 109
	KTs10            = 110
	KTs11            = 111
	KTs12            = 112
	KTs13            = 113
	KTs14            = 114
	KTs15            = 115
	KTs16            = 116
	KTs17            = 117
	KTs18            = 118
	KTs19            = 119
	KTs20            = 120
	KTs21            = 121
	KTs22            = 122
	KTs23            = 123
	KTs24            = 124
	KTs25            = 125
	KTs26            = 126
	KTs27            = 127
	KTs28            = 128
	KTs29            = 129
	KTs30            = 130
	KTs31            = 131
	KTs32            = 132
	KTs33            = 133
	KTs34            = 134
	KTs35            = 135
	KTs36            = 136
	KTs37            = 137
	KTs38            = 138
	KTs39            = 139
	KTs40            = 140
	KTs41            = 141
	KTs42            = 142
	KTs43            = 143
	KTs44            = 144
	KTs45            = 145
	KTs46            = 146
	KTs47            = 147
	KTs48            = 148
	KTs49            = 149
	KTs50            = 150
	KTs51            = 151
	KTs52            = 152
	KTs53            = 153
	KTs54            = 154
	KTs55            = 155
	KTs56            = 156
	KTs57            = 157
	KTs58            = 158
	KTs59            = 159
	KTs60            = 160
	KTs61            = 161
	KTs62            = 162
	KTs63            = 163
	KTs64            = 164
	KTs65            = 165
	KTs66            = 166
	KTs67            = 167
	KTs68            = 168
	KTs69            = 169
	KTs70            = 170
	KTs71            = 171
	KTs72            = 172
	KTs73            = 173
	KTs74            = 174
	KTs75            = 175
	KTs76            = 176
	KTs77            = 177
	KTs78            = 178
	KTs79            = 179
	KTs80            = 180
	KTs81            = 181
	KTs82            = 182
	KTs83            = 183
	KTs84            = 184
	KTs85            = 185
	KTs86            = 186
	KTs87            = 187
	KTs88            = 188
	KTs89            = 189
	KTs90            = 190
	KTs91            = 191
	KTs92            = 192
	KTs93            = 193
	KTs94            = 194
	KTs95            = 195
	KTs96            = 196
	KTs97            = 197
	KTs98            = 198
	KTs99            = 199
)


type KeyMap struct {
	Lang      string `json:"lang"`
	KBmargins struct {
		Top    float64 `json:"top"`
		Bottom float64 `json:"bottom"`
		Left   float64 `json:"left"`
		Right  float64 `json:"right"`
	} `json:"kbMargins"`
	TotalKeyWidth  float64 `json:"totalKeyWidth"`
	TotalRowHeight float64 `json:"totalRowHeight"`
	Rows           []struct {
		RowHeight float64 `json:"rowHeight"`
		Keys      []struct {
			IsPadding bool    `json:"isPadding"`
			KeyType   int     `json:"keyType"`
			KeyWidth  float64 `json:"keyWidth"`
			Char      string  `json:"char"`
		} `json:"keys"`
	} `json:"rows"`
}

type coords struct {
	X int
	Y int
}

// Key contains information about each key on the virtual keyboard
type Key struct {
	coord   coords
	width   int
	IsKey   bool
	KeyType int
	KeyCode []rune
}

// Row contains the height, and all the keys in the row.
type Row struct {
	rowHeight int
	keys      []Key
}

// VirtKeyboard contains the entire virtual keyboard, and the methods
// required to generate one from a keymap file
type VirtKeyboard struct {
	widthPX         int
	heightPX        int
	kmUnitWidth     int
	rhUnitWidth     int
	StartCoords     coords
	rows            []Row
	debounceStartTm time.Time
	prevKey         Key
}

// validateKeymap checks that the keymap file contains valid measurements
func validateKeymap(km *KeyMap) error {
	// Margin check
	if km.KBmargins.Bottom < 0 ||
		km.KBmargins.Top < 0 ||
		km.KBmargins.Left < 0 ||
		km.KBmargins.Right < 0 {
		return errors.New("keymap: negative numbers not allowed")
	} else if (km.KBmargins.Top+km.KBmargins.Bottom) > 0.8 &&
		(km.KBmargins.Left+km.KBmargins.Right) > 0.8 {
		return errors.New("combined margins exceed 0.8")
	}
	// Measurement check
	currRowMeas := float64(0)
	for i, r := range km.Rows {
		currRowMeas += r.RowHeight
		currKeyMeas := float64(0)
		for _, k := range r.Keys {
			currKeyMeas += k.KeyWidth
		}
		if currKeyMeas > km.TotalKeyWidth {
			errMsg := fmt.Sprintf("Key widths sum exceeds %f in row %d", km.TotalKeyWidth, i)
			return errors.New(errMsg)
		}
	}
	if currRowMeas > km.TotalRowHeight {
		errMsg := fmt.Sprintf("Row heights sum exceeds %f", km.TotalRowHeight)
		return errors.New(errMsg)
	}
	return nil
}

// New initilizes a VirtKeyboard for use
func New(km *KeyMap, fbWidth, fbHeight int) (*VirtKeyboard, error) {
	v := &VirtKeyboard{}
	if err := validateKeymap(km); err != nil {
		return v, err
	}
	// Calculate our margins in px from the percentages provided in the keymap
	floatFBw, floatFBh := float64(fbWidth), float64(fbHeight)
	pxFromTop := int(math.Round(floatFBh * km.KBmargins.Top))
	pxFromBot := int(math.Round(floatFBh * km.KBmargins.Bottom))
	pxFromLeft := int(math.Round(floatFBw * km.KBmargins.Left))
	pxFromRight := int(math.Round(floatFBw * km.KBmargins.Right))

	// Calculate our origin and dimensions from the margins
	v.StartCoords.X, v.StartCoords.Y = pxFromLeft, pxFromTop
	v.widthPX = fbWidth - pxFromLeft - pxFromRight
	v.heightPX = fbHeight - pxFromTop - pxFromBot

	// What's the width of each keymap unit? Rounded down to the nearest pixel of course
	v.kmUnitWidth = int(float64(v.widthPX) / km.TotalKeyWidth)
	// And the height of each rowheight unit?
	v.rhUnitWidth = int(float64(v.heightPX) / km.TotalRowHeight)
	v.prevKey.KeyCode = []rune{0}
	// time to give our keymap into a set of usable coordinates
	v.convertKeymap(km)
	return v, nil
}

// convertKeymap converts a keymap file into rows of keys with coordinate
// information
func (v *VirtKeyboard) convertKeymap(km *KeyMap) {
	currY := v.StartCoords.Y
	for _, r := range km.Rows {
		row := Row{}
		row.rowHeight = int(float64(v.rhUnitWidth) * r.RowHeight)
		ky := make([]Key, len(r.Keys))
		currX := v.StartCoords.X
		for j, k := range r.Keys {
			ky[j].width = int(float64(v.kmUnitWidth) * k.KeyWidth)
			ky[j].coord.Y = currY
			ky[j].coord.X = currX
			currX += ky[j].width
			ky[j].KeyType = k.KeyType
			ky[j].IsKey = !k.IsPadding
			if (ky[j].KeyType == 0 || ky[j].KeyType >= MaxKeyCodes) && len(k.Char) > 0 {
				runeSlice := []rune(k.Char)
				// We only care about the first rune...Except for programmable keys
				if ky[j].KeyType >= MaxKeyCodes {
					ky[j].KeyCode = runeSlice
				} else {
					ky[j].KeyCode = []rune{runeSlice[0]}
				}
			} else {
				ky[j].KeyCode = []rune{0}
			}
		}
		row.keys = ky
		currY += row.rowHeight
		v.rows = append(v.rows, row)
	}
}

// GetLabel returns a label for "special" keys
func (v *VirtKeyboard) GetLabel(kt int) string {
	switch kt {
	case KTalt:
		return "ALT"
	case KTbackspace:
		return "BKSP"
	case KTcapsLock:
		return "CAP"
	case KTcarriageReturn:
		return "RET"
	case KTcontrol:
		return "CTL"
	case KTdelete:
		return "DEL"
	case KTesc:
		return "ESC"
	case KTtab:
		return "TAB"
	case KTup:
		return "^"
	case KTdown:
		return "v"
	case KTleft:
		return "<"
	case KTright:
		return ">"
	case KTpgUp:
		return "<<"
	case KTpgDn:
		return ">>"
	case KTinsert:
		return "INS"
	case KThome:
		return "HOME"
	case KTend:
		return "END"
	case KTsym:
		return "SYM"
	case KTexit:
		return "EXIT"
	case KTf1:
		return "F1"
	case KTf2:
		return "F2"
	case KTf3:
		return "F3"
	case KTf4:
		return "F4"
	case KTf5:
		return "F5"
	case KTf6:
		return "F6"
	case KTf7:
		return "F7"
	case KTf8:
		return "F8"
	case KTf9:
		return "F9"
	case KTf10:
		return "F10"
	case KTf11:
		return "F11"
	case KTf12:
		return "F12"
	case KTf13:
		return "F13"
	case KTf14:
		return "F14"
	case KTf15:
		return "F15"
	case KTf16:
		return "F16"
	case KTf17:
		return "F17"
	case KTf18:
		return "F18"
	case KTf19:
		return "F19"
	case KTf20:
		return "F20"
	case KTs0:
		return "S0"
	case KTs1:
		return "S1"
	case KTs2:
		return "S2"
	case KTs3:
		return "S3"
	case KTs4:
		return "S4"
	case KTs5:
		return "S5"
	case KTs6:
		return "S6"
	case KTs7:
		return "S7"
	case KTs8:
		return "S8"
	case KTs9:
		return "S9"
	case KTs10:
		return "S10"
	case KTs11:
		return "S11"
	case KTs12:
		return "S12"
	case KTs13:
		return "S13"
	case KTs14:
		return "S14"
	case KTs15:
		return "S15"
	case KTs16:
		return "S16"
	case KTs17:
		return "S17"
	case KTs18:
		return "S18"
	case KTs19:
		return "S19"
	case KTs20:
		return "S20"
	case KTs21:
		return "S21"
	case KTs22:
		return "S22"
	case KTs23:
		return "S23"
	case KTs24:
		return "S24"
	case KTs25:
		return "S25"
	case KTs26:
		return "S26"
	case KTs27:
		return "S27"
	case KTs28:
		return "S28"
	case KTs29:
		return "S29"
	case KTs30:
		return "S30"
	case KTs31:
		return "S31"
	case KTs32:
		return "S32"
	case KTs33:
		return "S33"
	case KTs34:
		return "S34"
	case KTs35:
		return "S35"
	case KTs36:
		return "S36"
	case KTs37:
		return "S37"
	case KTs38:
		return "S38"
	case KTs39:
		return "S39"
	case KTs40:
		return "S40"
	case KTs41:
		return "S41"
	case KTs42:
		return "S42"
	case KTs43:
		return "S43"
	case KTs44:
		return "S44"
	case KTs45:
		return "S45"
	case KTs46:
		return "S46"
	case KTs47:
		return "S47"
	case KTs48:
		return "S48"
	case KTs49:
		return "S49"
	case KTs50:
		return "S50"
	case KTs51:
		return "S51"
	case KTs52:
		return "S52"
	case KTs53:
		return "S53"
	case KTs54:
		return "S54"
	case KTs55:
		return "S55"
	case KTs56:
		return "S56"
	case KTs57:
		return "S57"
	case KTs58:
		return "S58"
	case KTs59:
		return "S59"
	case KTs60:
		return "S60"
	case KTs61:
		return "S61"
	case KTs62:
		return "S62"
	case KTs63:
		return "S63"
	case KTs64:
		return "S64"
	case KTs65:
		return "S65"
	case KTs66:
		return "S66"
	case KTs67:
		return "S67"
	case KTs68:
		return "S68"
	case KTs69:
		return "S69"
	case KTs70:
		return "S70"
	case KTs71:
		return "S71"
	case KTs72:
		return "S72"
	case KTs73:
		return "S73"
	case KTs74:
		return "S74"
	case KTs75:
		return "S75"
	case KTs76:
		return "S76"
	case KTs77:
		return "S77"
	case KTs78:
		return "S78"
	case KTs79:
		return "S79"
	case KTs80:
		return "S80"
	case KTs81:
		return "S81"
	case KTs82:
		return "S82"
	case KTs83:
		return "S83"
	case KTs84:
		return "S84"
	case KTs85:
		return "S85"
	case KTs86:
		return "S86"
	case KTs87:
		return "S87"
	case KTs88:
		return "S88"
	case KTs89:
		return "S89"
	case KTs90:
		return "S90"
	case KTs91:
		return "S91"
	case KTs92:
		return "S92"
	case KTs93:
		return "S93"
	case KTs94:
		return "S94"
	case KTs95:
		return "S95"
	case KTs96:
		return "S96"
	case KTs97:
		return "S97"
	case KTs98:
		return "S98"
	case KTs99:
		return "S99"
	}
	return ""
}

// CreateIMG generates an image from the current keyboard.CreateIMG
// The current implementation saves the image as a PNG. This behaviour may
// change in the future to return an RBGA image
func (v *VirtKeyboard) CreateIMG(savePath, fontPath string) {
	kc := gg.NewContext(v.widthPX, v.heightPX)
	kc.DrawRectangle(0, 0, float64(v.widthPX), float64(v.heightPX))
	kc.SetRGB255(240, 240, 240)
	kc.Fill()
	kc.LoadFontFace(fontPath, 36)
	for _, r := range v.rows {
		for _, k := range r.keys {
			if k.IsKey {
				kc.Push()
				kx, ky := float64(k.coord.X-v.StartCoords.X), float64((k.coord.Y - v.StartCoords.Y))
				kw, kh := float64(k.width), float64(r.rowHeight)
				kmx, kmy := (kx + kw/2), (ky + kh/2)
				kc.DrawRectangle(kx, ky, kw, kh)
				kc.SetRGB255(0, 0, 0)
				kc.StrokePreserve()
				kc.SetRGB255(255, 255, 255)
				kc.Fill()
				kc.SetRGB255(0, 0, 0)
				if k.KeyType == KTstandardChar {
					kc.DrawStringAnchored(string(k.KeyCode), kmx, kmy, 0.5, 0.5)
				} else if (k.KeyType > KTstandardChar && k.KeyType < MaxKeyCodes) {
					kc.DrawStringAnchored(v.GetLabel(k.KeyType), kmx, kmy, 0.5, 0.5)
				} else if k.KeyType >= MaxKeyCodes {
					kc.DrawStringAnchored(string(k.KeyCode), kmx, kmy, 0.5, 0.5)
				}
				kc.Pop()
			}
		}
	}
	kc.SavePNG(savePath)
}

// GetPressedKey uses the coordinates provided to determine which
// key was pressed, and returns the key
func (v *VirtKeyboard) GetPressedKey(inX, inY int) (Key, error) {
	// Given X and Y, we need to find which key was pressed

	// First, reject any coordinates that are out of bounds
	if inY < v.StartCoords.Y || inY > (v.StartCoords.Y+v.heightPX) {
		return Key{}, errors.New("Y out of bounds")
	} else if inX < v.StartCoords.X || inX > (v.StartCoords.X+v.widthPX) {
		return Key{}, errors.New("Y out of bounds")
	}
	// Get the row index.
	rowIndex := -1
	currY := v.StartCoords.Y
	for i, r := range v.rows {
		if inY <= currY+r.rowHeight {
			rowIndex = i
			break
		}
		currY += r.rowHeight
	}
	// Getting key in row is a little trickier, as key width varies
	// Linear search, because our list will never be very large...
	if rowIndex >= 0 {
		keyNum := len(v.rows[rowIndex].keys)
		for i := 0; i < keyNum; i++ {
			k := v.rows[rowIndex].keys[i]
			if inX <= (k.coord.X + k.width) {
				if (k.KeyType > KTstandardChar && k.KeyType != v.prevKey.KeyType) || (k.KeyType == KTstandardChar && k.KeyCode[0] != v.prevKey.KeyCode[0]) {
					v.prevKey = k
					v.debounceStartTm = time.Now()
					return k, nil
				} else {
					if time.Since(v.debounceStartTm) < (200 * time.Millisecond) {
						return Key{}, errors.New("debounce detected")
					} else {
						v.prevKey = k
						v.debounceStartTm = time.Now()
						return k, nil
					}
				}
			}
		}
	}
	return Key{}, errors.New("key not found")
}
